<?php

namespace Drupal\hierarchical_taxonomy_importer\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\hierarchical_taxonomy_importer\services\ImporterService;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\File\FileSystemInterface;

/**
 * Provides a form for importing hierarchical taxonomy terms from CSV.
 */
class TaxonomyImporterForm extends FormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The hierarchical taxonomy importer service.
   *
   * @var \Drupal\hierarchical_taxonomy_importer\services\ImporterService
   */
  protected $hierarchicalTaxonomyImporter;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Constructs a new TaxonomyImporterForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\hierarchical_taxonomy_importer\services\ImporterService $hierarchical_taxonomy_importer_importer
   *   The hierarchical taxonomy importer service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ImporterService $hierarchical_taxonomy_importer_importer,
    MessengerInterface $messenger,
    FileSystemInterface $file_system,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->hierarchicalTaxonomyImporter = $hierarchical_taxonomy_importer_importer;
    $this->messenger = $messenger;
    $this->fileSystem = $file_system;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('hierarchical_taxonomy_importer.importer'),
      $container->get('messenger'),
      $container->get('file_system')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'taxonomy_importer_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['vocabulary'] = [
      '#type' => 'select',
      '#title' => $this->t('Vocabularies'),
      '#description' => $this->t('Select a vocabulary to import taxonomies terms from CSV file.'),
      '#options' => $this->getVocabularies(),
      '#required' => TRUE,
      '#size' => 1,
    ];
    $form['csv_file'] = [
      '#type' => 'file',
      '#title' => $this->t('CSV File'),
      '#description' => $this->t('Upload a CSV file with taxonomy information.'),
      '#required' => TRUE,
    ];
    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Import'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    // Getting file array.
    $files = $this->getRequest()->files->get('files', []);

    // Checking if file is uploaded.
    if (!empty($files['csv_file'])) {
      $file_upload = $files['csv_file'];
      if ($file_upload->isValid()) {
        $form_state->setValue('csv_file', $file_upload->getRealPath());

        // Get the file extension instead of relying on MIME type.
        $file_name = $file_upload->getClientOriginalName();
        $extension = pathinfo($file_name, PATHINFO_EXTENSION);

        // Check if the file extension is csv.
        if (strtolower($extension) !== 'csv') {
          $form_state->setErrorByName('csv_file', $this->t('Please upload a file with .csv extension.'));
        }

        return;
      }
    }

    $form_state->setErrorByName('csv_file', $this->t('The file could not be uploaded.'));
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $csv_file_path = $form_state->getValue('csv_file');

    if (($handle = fopen($csv_file_path, "r")) !== FALSE) {
      $vid = $form_state->getValue('vocabulary');
      $output = [];

      while (($data = fgetcsv($handle, 1000, ",")) !== FALSE) {
        $output[] = $data;
      }

      fclose($handle);

      try {
        $this->hierarchicalTaxonomyImporter->import($vid, $output);
        $this->messenger->addStatus($this->t('All terms have been imported successfully.'));
      }
      catch (\Exception $e) {
        $this->messenger->addError($this->t('An error occurred during import: @error', ['@error' => $e->getMessage()]));
      }
    }
    else {
      $this->messenger->addError($this->t('Could not open the CSV file for reading.'));
    }
  }

  /**
   * Returns the list of existing vocabularies.
   *
   * @return array
   *   List of existing Vocabularies in an associative array's form.
   */
  public function getVocabularies() {
    $output = [];

    // Loading existing vocabularies.
    $vocabularies = Vocabulary::loadMultiple();

    // If vocabularies are not empty then load them to an array one by one
    // Vocabulary ID would be used as offset and label as value of array on that
    // offset.
    if (!empty($vocabularies)) {
      // Traversal of vocabularies is taking place
      // and preparing an options array.
      foreach ($vocabularies as $vocabulary) {
        $output[$vocabulary->id()] = $vocabulary->label();
      }
    }

    // Return the output back to the form.
    return $output;
  }

}
