<?php

namespace Drupal\hierarchical_taxonomy_importer\Importer;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\hierarchical_taxonomy_importer\Base\ImporterBase;
use Drupal\taxonomy\Entity\Term;

/**
 * Provides functionality to import terms from CSV into hierarchical taxonomies.
 */
class CsvImporter extends ImporterBase {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Counter to track the current position in the data array.
   *
   * @var int
   */
  protected $counter;

  /**
   * Constructs a CsvImporter object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct();
    $this->entityTypeManager = $entity_type_manager;
    $this->counter = 0;
  }

  /**
   * Sets the current counter value.
   *
   * @param int $counter
   *   The counter value to set.
   */
  private function setCounter(int $counter): void {
    $this->counter = $counter;
  }

  /**
   * Gets the current counter value.
   *
   * @return int
   *   The current counter value.
   */
  private function getCounter(): int {
    return $this->counter;
  }

  /**
   * Recursively processes taxonomy terms to build a hierarchical structure.
   *
   * This method handles the creation of taxonomy terms with appropriate
   * parent-child relationships based on the provided data array.
   *
   * @param string $vid
   *   The vocabulary ID.
   * @param array $data
   *   The array of term data to process.
   * @param int $parent
   *   The parent term ID, defaults to 0 for top-level terms.
   *
   * @return mixed
   *   The term ID of the last processed term or the value from the data array.
   */
  public function getNestedTid(string $vid, array $data, int $parent = 0) {
    // Current term name.
    $term = $data[0];

    // Handle single-element data array (top-level term).
    if (count($data) == 1 && empty($data[1])) {
      // Create a new top-level term.
      $parent_new = Term::create([
        'name' => $data[0],
        'vid' => $vid,
      ]);
      $parent_new->save();
      $this->counter++;
      return $parent_new->id();
    }

    // Handle hierarchical terms.
    if (!empty($data[$this->counter]) && !empty($data[$this->counter + 1])) {
      // Find the parent term by name.
      $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
      $parent_terms = $term_storage->loadByProperties([
        'name' => $data[$this->counter + 1],
        'vid' => $vid,
      ]);

      if (!empty($parent_terms)) {
        // Create a new term with the found parent.
        $parent_id = reset($parent_terms)->id();
        $new_term = Term::create([
          'name' => $term,
          'parent' => [$parent_id],
          'vid' => $vid,
        ]);
        $new_term->save();

        // Process next level in hierarchy.
        $this->counter++;
        return $this->getNestedTid($vid, $data, $parent_id);
      }
      else {
        // Parent term not found, create it first.
        $parent_term = Term::create([
          'name' => $data[$this->counter + 1],
          'vid' => $vid,
        ]);
        $parent_term->save();
        $parent_id = $parent_term->id();

        // Create the child term.
        $new_term = Term::create([
          'name' => $term,
          'parent' => [$parent_id],
          'vid' => $vid,
        ]);
        $new_term->save();

        // Process next level in hierarchy.
        $this->counter++;
        return $this->getNestedTid($vid, $data, $parent_id);
      }
    }
    else {
      // End of the hierarchy for this branch.
      return $data[$this->counter] ?? NULL;
    }
  }

  /**
   * Imports taxonomy terms from provided data.
   *
   * @param string $vid
   *   The vocabulary ID.
   * @param array $data
   *   The term data to import.
   * @param bool $reset
   *   Whether to reset existing terms.
   *
   * @throws \Exception
   *   Throws an exception if the vocabulary ID is invalid.
   */
  public function importTaxonomies(string $vid, array $data, bool $reset = FALSE): void {
    // Checking for a valid VID.
    if (empty($vid)) {
      throw new \Exception("Vocabulary ID is invalid or null");
    }

    // Reset counter for each import operation.
    $this->counter = 0;

    // Process the taxonomy data.
    $this->getNestedTid($vid, $data, 0);
  }

}
