<?php

namespace Drupal\hierarchical_taxonomy_importer\services;

use Drupal\taxonomy\Entity\Term;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Provides a service for importing hierarchical taxonomy terms.
 */
class ImporterService {

  /**
   * Constant for handling terms with the same parent.
   */
  const SAME_PARENT = 0;

  /**
   * Constant for handling terms with a different parent.
   */
  const DIFFERENT_PARENT = 1;

  /**
   * Constant for handling terms when moving up the hierarchy.
   */
  const PREVIOUS_PARENT = -1;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Constructs a new ImporterService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    Connection $connection,
    LoggerChannelFactoryInterface $logger_factory,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->connection = $connection;
    $this->logger = $logger_factory->get('hierarchical_taxonomy_importer');
  }

  /**
   * Imports taxonomy terms from an array structure.
   *
   * Recursively processes the data array to create taxonomy terms with proper
   * hierarchical relationships.
   *
   * @param string $vid
   *   Vocabulary ID to import terms.
   * @param array $data
   *   Array structure with values to import as terms.
   * @param int $count
   *   Counter for total number of records checked to stop the execution.
   * @param int $row
   *   Current row number for CSV row.
   * @param int $pointer
   *   Current Column Pointer.
   * @param int $parent
   *   Parent TID for term being created.
   * @param int $tag
   *   Flag used for checking if there is a record on next level,
   *   same level or on top level.
   *
   * @return int|null
   *   Returns the parent term ID or NULL.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function import(string $vid, array $data, int $count = 0, int $row = 0, int $pointer = 0, int $parent = 0, int $tag = 1): ?int {
    // If all rows have been traversed then exit.
    if ($count >= count($data)) {
      return NULL;
    }

    // If all levels of a parent have been traversed
    // and new parent has come up for
    // import then reset the levels counter back to 0 for hierarchical connect.
    if (!empty($data[$row][0])) {
      $pointer = 0;
      $parent = 0;
    }

    // Get the index of non-null value in the current row.
    $pointer = $this->getIndexOfNonNullValues($data[$row]);

    // If pointer is null then return null.
    if (is_null($pointer)) {
      return NULL;
    }

    // Current Column value.
    $term_name = $data[$row][$pointer];

    // Initialize parent term name.
    $parent_term_name = "";

    // Find the parent row for the current record.
    $parent_row = $this->getParentRow($data, $row, $pointer);

    // Process parent relationships if we have a valid parent row and pointer.
    if ($parent_row >= 0 && $pointer > 0) {
      $parent_term_name = $data[$parent_row][$pointer - 1];

      // Get the hierarchical path to the current term.
      $csv_tree = $this->getParentTree($data, $row, $pointer);

      // If parent term exists, find the correct parent ID.
      if (!empty($parent_term_name)) {
        // Load all terms with the same name as the parent in csv.
        $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
        $matching_parents = $term_storage->loadByProperties([
          'vid' => $vid,
          'name' => $parent_term_name,
        ]);

        $parent_tree = [];

        // Get all parents with the same name as the supposed
        // parent and construct a list of term names.
        foreach ($matching_parents as $mp) {
          $term_list = $term_storage->loadAllParents($mp->id());
          foreach ($term_list as $term) {
            $parent_tree[$mp->id()][] = $term->getName();
          }
        }

        // Find matching parent based on the hierarchy.
        foreach ($parent_tree as $parent_id => $children_list) {
          if ($parent_tree[$parent_id] == $csv_tree) {
            $parent = $parent_id;
          }
        }
      }
    }

    // Store original parent for sibling terms.
    $original_parent = $parent;

    // Create the current term and update the parent ID as needed.
    $parent = $this->updateActualParent($vid, $term_name, $parent, $tag);

    // Process next level term if it exists.
    if (!empty($data[$row + 1][$pointer + 1])) {
      return $this->import($vid, $data, $count + 1, $row + 1, $pointer + 1, $parent, self::DIFFERENT_PARENT);
    }

    // Process sibling term if it exists.
    if (!empty($data[$row + 1][$pointer])) {
      return $this->import($vid, $data, $count + 1, $row + 1, $pointer, $original_parent, self::SAME_PARENT);
    }

    // If no siblings or children, move back up the hierarchy.
    return $this->import($vid, $data, $count + 1, $row + 1, $pointer - 1, $parent, self::PREVIOUS_PARENT);
  }

  /**
   * Returns the index of non-null values in an array row.
   *
   * @param array|null $data
   *   Row data from the CSV.
   *
   * @return int|null
   *   The column index of the first non-empty value, or NULL if none found.
   */
  public function getIndexOfNonNullValues(?array $data): ?int {
    // This gives the index of column in CSV that would be created as a new
    // taxonomy term.
    if (!empty($data) && !is_null($data)) {
      // Return the column after removing null or blank values from a row.
      $keys = array_keys(array_diff($data, [" ", ""]));
      return !empty($keys) ? array_shift($keys) : 0;
    }
    return 0;
  }

  /**
   * Returns all parents of current term from the CSV data.
   *
   * Builds a hierarchical path from the current term to the root.
   *
   * @param array $data
   *   The CSV data array.
   * @param int $row
   *   The current row index.
   * @param int $pointer
   *   The current column pointer.
   *
   * @return array
   *   An array of parent term names in hierarchical order.
   */
  public function getParentTree(array $data, int $row, int $pointer): array {
    $tree = [];
    $iterator = $row;
    $level = $pointer - 1;

    while ($level >= 0 && $iterator >= 0) {
      if (!empty($data[$iterator][$level])) {
        $tree[] = $data[$iterator][$level];
        $level -= 1;
      }
      $iterator -= 1;
    }

    return $tree;
  }

  /**
   * Finds the row index of the parent term.
   *
   * @param array $data
   *   CSV Data as an array.
   * @param int $parent_row
   *   Starting row index to search from.
   * @param int $pointer
   *   Current column pointer.
   *
   * @return int
   *   The row index of the parent term.
   */
  protected function getParentRow(array $data, int $parent_row, int $pointer): int {
    // Look upwards in the CSV for the parent term.
    while (empty($data[$parent_row][$pointer - 1]) && $parent_row > 0) {
      $parent_row = $parent_row - 1;
    }

    return $parent_row;
  }

  /**
   * Creates a new taxonomy term.
   *
   * @param string $vid
   *   Vocabulary ID.
   * @param string $term
   *   Term name.
   * @param int $parent
   *   Parent Term ID, 0 for root level.
   *
   * @return int|null
   *   The ID of the newly created term, or NULL on failure.
   */
  protected function createNewTerm(string $vid, string $term, int $parent): ?int {
    try {
      // Create a new taxonomy term with proper parent relationship.
      $new_term = Term::create([
        'name' => $term,
        'parent' => $parent > 0 ? [$parent] : [],
        'vid' => $vid,
      ]);
      $new_term->save();

      return $new_term->id();
    }
    catch (\Exception $ex) {
      $this->logger->error('Error creating taxonomy term: @message', ['@message' => $ex->getMessage()]);
      return NULL;
    }
  }

  /**
   * Updates the parent ID based on the term relationship type.
   *
   * @param string $vid
   *   Vocabulary ID.
   * @param string $term_name
   *   Current taxonomy term name.
   * @param int $parent
   *   Parent term ID.
   * @param int $tag
   *   Flag indicating relationship type (same parent, different parent, etc.).
   *
   * @return int
   *   The updated parent ID.
   */
  protected function updateActualParent(string $vid, string $term_name, int $parent, int $tag): int {
    $new_term_id = $this->createNewTerm($vid, $term_name, $parent);

    // Only update the parent reference if this isn't a sibling term.
    if ($tag !== self::SAME_PARENT && !is_null($new_term_id)) {
      $parent = $new_term_id;
    }

    return $parent ?? 0;
  }

}
